@@ -7,7 +7,9 @@ import os
 from fun import (
 					msb_size,
 					stream_copy, 
-					apply_delta_data, 
+					apply_delta_data,
+					connect_deltas,
+					DeltaChunkList,
 					delta_types
 				)
 
@@ -20,6 +22,13 @@ from util import (
 		zlib
 	)
 
+has_perf_mod = False
+try:
+	from _perf import apply_delta as c_apply_delta
+	has_perf_mod = True
+except ImportError:
+	pass
+
 __all__ = ('DecompressMemMapReader', 'FDCompressedSha1Writer', 'DeltaApplyReader')
 
 
@@ -309,6 +318,10 @@ class DeltaApplyReader(LazyMixin):
 					"_br"					# number of bytes read 
 				)
 	
+	#{ Configuration
+	k_max_memory_move = 250*1000*1000
+	#} END configuration
+	
 	def __init__(self, stream_list):
 		"""Initialize this instance with a list of streams, the first stream being 
 		the delta to apply on top of all following deltas, the last stream being the
@@ -319,10 +332,47 @@ class DeltaApplyReader(LazyMixin):
 		self._dstreams = tuple(stream_list[:-1])
 		self._br = 0
 		
-	def _set_cache_(self, attr):
+	def _set_cache_too_slow_without_c(self, attr):
+		# the direct algorithm is fastest and most direct if there is only one 
+		# delta. Also, the extra overhead might not be worth it for items smaller
+		# than X - definitely the case in python, every function call costs 
+		# huge amounts of time
+		# if len(self._dstreams) * self._bstream.size < self.k_max_memory_move:
+		if len(self._dstreams) == 1:
+			return self._set_cache_brute_(attr)
+		
+		# Aggregate all deltas into one delta in reverse order. Hence we take 
+		# the last delta, and reverse-merge its ancestor delta, until we receive
+		# the final delta data stream.
+		# print "Handling %i delta streams, sizes: %s" % (len(self._dstreams), [ds.size for ds in self._dstreams])
+		dcl = connect_deltas(self._dstreams)
+		
+		# call len directly, as the (optional) c version doesn't implement the sequence
+		# protocol
+		if dcl.rbound() == 0:
+			self._size = 0
+			self._mm_target = allocate_memory(0)
+			return
+		# END handle empty list
+		
+		self._size = dcl.rbound()
+		self._mm_target = allocate_memory(self._size)
+		
+		bbuf = allocate_memory(self._bstream.size)
+		stream_copy(self._bstream.read, bbuf.write, self._bstream.size, 256 * mmap.PAGESIZE)
+		
+		# APPLY CHUNKS
+		write = self._mm_target.write
+		dcl.apply(bbuf, write)
+		
+		self._mm_target.seek(0)
+		
+	def _set_cache_brute_(self, attr):
 		"""If we are here, we apply the actual deltas"""
 		
-		# prefetch information
+		# TODO: There should be a special case if there is only one stream
+		# Then the default-git algorithm should perform a tad faster, as the
+		# delta is not peaked into, causing less overhead.
 		buffer_info_list = list()
 		max_target_size = 0
 		for dstream in self._dstreams:
@@ -373,7 +423,10 @@ class DeltaApplyReader(LazyMixin):
 			stream_copy(dstream.read, ddata.write, dstream.size, 256*mmap.PAGESIZE)
 			
 			#######################################################################
-			apply_delta_data(bbuf, src_size, ddata, len(ddata), tbuf)
+			if 'c_apply_delta' in globals():
+				c_apply_delta(bbuf, ddata, tbuf);
+			else:
+				apply_delta_data(bbuf, src_size, ddata, len(ddata), tbuf.write)
 			#######################################################################
 			
 			# finally, swap out source and target buffers. The target is now the 
@@ -390,12 +443,14 @@ class DeltaApplyReader(LazyMixin):
 		self._mm_target = bbuf
 		self._size = final_target_size
 		
-		# TODO: Once that works, figure out the ordering of the opcodes. If they
-		# are always in-order/sequential, an alternate implementation could 
-		# use stream access only. Of course this would mean we would read 
-		# all deltas in advance, analyse the opcode ranges to determine a final
-		# concatenated opcode list which indicates what to copy from which delta
-		# to which position. This preprocessing would allow true streaming
+	
+	#{ Configuration
+	if not has_perf_mod:
+		_set_cache_ = _set_cache_brute_
+	else:
+		_set_cache_ = _set_cache_too_slow_without_c
+	
+	#} END configuration
 		
 	def read(self, count=0):
 		bl = self._size - self._br		# bytes left
@@ -613,4 +668,7 @@ class NullStream(object):
 	def write(self, data):
 		return len(data)
 
+
 #} END W streams
+
+
